from typing import Any

from apidaora import (
    MiddlewareRequest,
    Middlewares,
    Response,
    appdaora,
    route,
    text,
)


def request_extra_args(request: MiddlewareRequest) -> None:
    request.kwargs = {'extra': 'You'}


def response_extra_args(
    request: MiddlewareRequest, response: Response
) -> None:
    if request.kwargs and response.kwargs:
        response.body = response.body.replace(
            'You', f"{request.kwargs['extra']} and {response.kwargs['name']}"
        )


@route.get(
    '/middlewares-kwargs',
    middlewares=Middlewares(
        pre_execution=[request_extra_args],
        post_execution=[response_extra_args],
    ),
)
async def extra_args_controller(name: str, **kwargs: Any) -> Response:
    return text(hello(kwargs['extra']), name=name)


def hello(name: str) -> str:
    return f'Hello {name}!'


app = appdaora(extra_args_controller)



from collections import defaultdict
from typing import Any, DefaultDict, List, Optional, Sequence, Type, Union

from .asgi.app import asgi_app
from .asgi.base import ASGIApp
from .asgi.router import Controller, Route, make_router
from .class_controller import ClassController
from .controllers.background_task import BackgroundTask
from .method import MethodType
from .middlewares import Middlewares
from .options import make_options_controller
from .route.factory import make_route


Controllers = Union[
    Controller,
    BackgroundTask,
    ClassController,
    Type[ClassController],
    Sequence[
        Union[
            Controller, BackgroundTask, ClassController, Type[ClassController],
        ]
    ],
    Any,
]


def appdaora(
    controllers: Controllers,
    middlewares: Optional[Middlewares] = None,
    options: bool = False,
) -> ASGIApp:
    routes = []
    func_controllers: List[Union[Controller, BackgroundTask]] = []
    path_methods_map: DefaultDict[str, List[MethodType]] = defaultdict(list)

    def update_path_methods_map(routes: Sequence[Route]) -> None:
        for route in routes:
            path_methods_map[route.path_pattern].append(route.method)

    def update_path_methods_map_with_route(routes: Sequence[Route]) -> None:
        for route in routes:
            if route.has_options:
                path_methods_map[route.path_pattern].append(route.method)

    if not isinstance(controllers, Sequence):
        controllers = [controllers]

    for controller in controllers:
        if (
            isinstance(controller, ClassController)
            or (
                isinstance(controller, type)
                and issubclass(controller, ClassController)
            )
            or (hasattr(controller, 'path'))
        ):
            routes.extend(get_class_routes(controller))
        else:
            func_controllers.append(controller)  # type: ignore

    for controller in func_controllers:
        if isinstance(controller, BackgroundTask):
            routes.extend(controller.create.routes)
            routes.extend(controller.get_results.routes)
            if options:
                update_path_methods_map(controller.create.routes)
                update_path_methods_map(controller.get_results.routes)
            else:
                update_path_methods_map_with_route(controller.create.routes)
                update_path_methods_map_with_route(
                    controller.get_results.routes
                )

        else:
            routes.extend(controller.routes)
            if options:
                update_path_methods_map(controller.routes)
            else:
                update_path_methods_map_with_route(controller.routes)

        for path, methods in path_methods_map.items():
            controller_ = make_options_controller(methods)
            routes.append(
                make_route(
                    path,
                    MethodType.OPTIONS,
                    controller_,
                    has_content_length=False,
                )
            )

    return asgi_app(make_router(routes, middlewares=middlewares))


def get_class_routes(controller: Any) -> List[Route]:
    if isinstance(controller, type):
        controller = controller()

    return [
        make_route(
            path_pattern=controller.path,
            method=method,
            controller=getattr(controller, method_str),
        )
        for method in MethodType
        if hasattr(controller, method_str := method.value.lower())  # noqa
    ]



from enum import Enum


class MethodType(Enum):
    GET = 'GET'
    POST = 'POST'
    PATCH = 'PATCH'
    PUT = 'PUT'
    DELETE = 'DELETE'
    OPTIONS = 'OPTIONS'
    HEAD = 'HEAD'
    TRACE = 'TRACE'
    CONNECT = 'CONNECT'



from dataclasses import dataclass
from typing import Any, Callable, ClassVar, Dict, Type

from dictdaora import DictDaora
from jsondaora import jsondaora

from .annotations_getters import (
    get_annotations_body,
    get_annotations_headers,
    get_annotations_path_args,
    get_annotations_query_dict,
)


@dataclass
class AnnotationsInfo:
    has_input: bool = False
    has_path_args: bool = False
    has_query_dict: bool = False
    has_headers: bool = False
    has_body: bool = False
    has_kwargs: bool = False


class ControllerInput(DictDaora):
    __annotations_info__: ClassVar[AnnotationsInfo] = AnnotationsInfo()
    __annotations_path_args__: ClassVar[Dict[str, Type[Any]]] = {}
    __annotations_query_dict__: ClassVar[Dict[str, Type[Any]]] = {}
    __annotations_headers__: ClassVar[Dict[str, Type[Any]]] = {}
    __annotations_body__: ClassVar[Dict[str, Type[Any]]] = {}
    __headers_name_map__: ClassVar[Dict[str, str]] = {}


def controller_input(
    controller: Callable[..., Any], path_pattern: str
) -> Type[ControllerInput]:
    annotations_info = AnnotationsInfo()
    headers_name_map: Dict[str, str] = {}

    if hasattr(controller, '__annotations__'):
        annotations_path_args = get_annotations_path_args(
            path_pattern, controller.__annotations__
        )
        annotations_headers = get_annotations_headers(
            headers_name_map, controller.__annotations__
        )
        annotations_query_dict = get_annotations_query_dict(
            annotations_path_args,
            annotations_headers,
            controller.__annotations__,
        )
        annotations_body = get_annotations_body(controller.__annotations__)
        all_annotations = (
            annotations_path_args,
            annotations_query_dict,
            annotations_headers,
            annotations_body,
        )
        annotations = {
            name: type_
            for partial_annotations in all_annotations
            for name, type_ in partial_annotations.items()
        }

        if annotations:
            annotations_info.has_input = True

            if annotations_path_args:
                annotations_info.has_path_args = True

            if annotations_query_dict:
                annotations_info.has_query_dict = True

            if annotations_headers:
                annotations_info.has_headers = True

            if annotations_body:
                annotations_info.has_body = True

            if 'kwargs' in controller.__annotations__:
                annotations_info.has_kwargs = True

            AnnotatedControllerInput = jsondaora(
                type(
                    'AnnotatedControllerInput',
                    (ControllerInput,),
                    {
                        '__annotations__': annotations,
                        '__annotations_info__': annotations_info,
                        '__annotations_path_args__': annotations_path_args,
                        '__annotations_query_dict__': annotations_query_dict,
                        '__annotations_headers__': annotations_headers,
                        '__annotations_body__': annotations_body,
                        '__headers_name_map__': headers_name_map,
                    },
                )
            )
            return AnnotatedControllerInput  # type: ignore

    return ControllerInput



from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List, Optional, Set, Type

from .exceptions import BadRequestError
from .header import Header
from .responses import Response
from .route.controller_input import ControllerInput


@dataclass
class MiddlewareRequest:
    path_pattern: str
    path_args: Optional[Dict[str, Any]] = None
    query_dict: Optional[Dict[str, Any]] = None
    headers: Optional[Dict[str, Header]] = None
    body: Any = None
    kwargs: Optional[Dict[str, Any]] = None


@dataclass
class Middlewares:
    post_routing: List[Callable[[str, Dict[str, Any]], None]] = field(
        default_factory=list
    )
    pre_execution: List[Callable[[MiddlewareRequest], None]] = field(
        default_factory=list
    )
    post_execution: List[
        Callable[[MiddlewareRequest, Response], None]
    ] = field(default_factory=list)


def make_asgi_input_from_requst(
    request: MiddlewareRequest, input_cls: Type[ControllerInput]
) -> ControllerInput:
    kwargs = {}

    if request.path_args:
        kwargs.update(request.path_args)
    if request.query_dict:
        kwargs.update(request.query_dict)
    if request.headers:
        kwargs.update(request.headers)
    if request.body:
        kwargs['body'] = request.body
    if request.kwargs and (input_cls.__annotations_info__.has_kwargs):
        kwargs.update(request.kwargs)

    return input_cls(**kwargs)


class AllowOriginHeader(
    Header, type=str, http_name='access-control-allow-origin'
):
    ...


class ExposeHeadersHeader(
    Header, type=str, http_name='access-control-expose-headers'
):
    ...


class AllowHeadersHeader(
    Header, type=str, http_name='access-control-allow-headers'
):
    ...


class AllowMethodsHeader(
    Header, type=str, http_name='access-control-allow-methods'
):
    ...


class CorsMiddleware:
    def __init__(
        self,
        *,
        allow_origin: Optional[str] = None,
        expose_headers: Optional[str] = None,
        allow_headers: Optional[str] = None,
        allow_methods: Optional[str] = None,
        servers_all: Optional[str] = '*',
    ):
        if (
            not allow_origin
            and not expose_headers
            and not allow_headers
            and not allow_methods
        ):
            self.headers = [
                AllowOriginHeader(servers_all),
                ExposeHeadersHeader(servers_all),
                AllowHeadersHeader(servers_all),
                AllowMethodsHeader(servers_all),
            ]
        else:
            self.headers = []
            if allow_origin:
                self.headers.append(AllowOriginHeader(allow_origin))
            if expose_headers:
                self.headers.append(ExposeHeadersHeader(expose_headers))
            if allow_headers:
                self.headers.append(AllowHeadersHeader(allow_headers))
            if allow_methods:
                self.headers.append(AllowMethodsHeader(allow_methods))

        self.headers_tuple = tuple(self.headers)

    def __call__(self, request: MiddlewareRequest, response: Response) -> None:
        if isinstance(response.headers, list):
            response.headers.extend(self.headers)

        elif isinstance(response.headers, tuple):
            response.headers = response.headers + self.headers_tuple

        elif response.headers is None:
            response.headers = self.headers_tuple


@dataclass
class LockRequestMiddleware:
    locks: Set[str] = field(default_factory=set)

    def lock(self, path_pattern: str, path_args: Dict[str, Any]) -> None:
        if path_pattern in self.locks:
            raise BadRequestError(
                'lock-request', {'path_pattern': path_pattern}
            )

        self.locks.add(path_pattern)

    def unlock_pre_execution(self, request: MiddlewareRequest) -> None:
        self.locks.remove(request.path_pattern)

    def unlock_post_execution(
        self, request: MiddlewareRequest, response: Response
    ) -> None:
        self.locks.remove(request.path_pattern)
